// This is free and unencumbered software released into the public
// domain.  For more information, see <http://unlicense.org> or the
// accompanying UNLICENSE file.

package pers_test

import (
	"fmt"
	"testing"

	"git.sr.ht/~nelsam/hel/pkg/pers"
	"git.sr.ht/~nelsam/hel/vegr"
	"git.sr.ht/~nelsam/hel/vegr/ret"
	"github.com/poy/onpar"
	"github.com/poy/onpar/expect"
)

func TestReturn(t *testing.T) {
	o := onpar.BeforeEach(onpar.New(t), func(t *testing.T) expectation {
		return expect.New(t)
	})
	defer o.Run()

	type entry struct {
		returnTo any
		values   []any
		panicMsg string
	}
	onpar.TableSpec(o, func(expect expectation, test entry) {
		defer func() {
			r := recover()
			expect(r).To(not(beNil()))
			expect(fmt.Sprintf("%v", r)).To(equal(test.panicMsg))
		}()
		pers.Return(test.returnTo, test.values...)
	}).
		Entry("it panics with an empty struct", entry{
			returnTo: struct{}{},
			panicMsg: "pers: returning [] on mock (struct {}) is not possible: sel: empty struct (struct {}) is not a valid mock field",
		}).
		Entry("it panics with an invalid struct", entry{
			returnTo: struct{ Foo int }{},
			values:   []any{1},
			panicMsg: "pers: returning [1] on mock (struct { Foo int }) is not possible: sel: field 0 of type struct { Foo int } cannot be used as a mock field: sel: type int is neither a struct of channels nor a channel",
		}).
		Entry("it panics with too few arguments", entry{
			returnTo: struct {
				Foo chan int
				Bar chan string
			}{
				Foo: make(chan int, 1),
				Bar: make(chan string, 1),
			},
			values:   []any{1},
			panicMsg: "pers: returning [1] on mock (struct { Foo chan int; Bar chan string }) is not possible: sel: expected 2 arguments for mock (struct { Foo chan int; Bar chan string }); got 1",
		}).
		Entry("it panics with too many arguments", entry{
			returnTo: make(chan int, 1),
			values:   []any{2, "foo"},
			panicMsg: "pers: returning [2 foo] on mock (chan int) is not possible: sel: expected 1 argument for mock (chan int); got 2",
		}).
		Entry("it panics with a channel that would block", entry{
			returnTo: struct {
				Foo chan int
				Bar chan string
			}{
				Foo: make(chan int),
				Bar: make(chan string),
			},
			values:   []any{1, "foo"},
			panicMsg: "pers: returning [1 foo] on mock (struct { Foo chan int; Bar chan string }) would block [hint: increase channel buffer size]",
		})

	o.Spec("it handles nil values correctly", func(expect expectation) {
		c := make(chan error, 1)
		pers.Return(c, nil)
		expect(c).To(chain(receive(), equal(nil)))
	})

	o.Spec("it returns on a channel", func(expect expectation) {
		c := make(chan int, 1)
		pers.Return(c, 1)
		expect(c).To(chain(receive(), equal(1)))
	})

	o.Spec("it returns on a struct of channels", func(expect expectation) {
		type fooReturns struct {
			Foo chan string
			Bar chan bool
		}
		v := fooReturns{make(chan string, 1), make(chan bool, 1)}
		pers.Return(v, "foo", true)
		expect(v.Foo).To(chain(receive(), equal("foo")))
		expect(v.Bar).To(chain(receive(), equal(true)))
	})

	o.Spec("it skips blocking channels", func(expect expectation) {
		type fooReturns struct {
			Foo ret.Blocker
			Bar chan bool
		}
		v := fooReturns{Bar: make(chan bool, 1)}
		defer func() {
			if r := recover(); r != nil {
				t.Fatalf("panic caught; pers.Return probably did not skip the blocker field: %v", r)
			}
		}()
		pers.Return(v, true)
		expect(v.Bar).To(chain(receive(), equal(true)))
	})

	type customString string

	o.Spec("it converts convertible values", func(expect expectation) {
		type fooReturns struct {
			Foo chan customString
			Bar chan int64
		}
		v := fooReturns{make(chan customString, 1), make(chan int64, 1)}
		pers.Return(v, "foo", 7)
		expect(v.Foo).To(chain(receive(), equal(customString("foo"))))
		expect(v.Bar).To(chain(receive(), equal(int64(7))))
	})

	o.Spec("it returns nil on panicker channels", func(expect expectation) {
		type fooReturns struct {
			Foo ret.Blocker
			Bar ret.Panicker
			Baz chan bool
		}
		v := fooReturns{Foo: vegr.BlockChan(), Bar: make(ret.Panicker, 1), Baz: make(chan bool, 1)}
		pers.Return(v, true)
		expect(v.Bar).To(chain(receive(), beNil()))
		expect(v.Baz).To(chain(receive(), equal(true)))
	})
}
